/*
 * INET		An implementation of the TCP/IP protocol suite for the LINUX
 *		operating system.  INET is implemented using the  BSD Socket
 *		interface as the means of communication with the user level.
 *
 *		Internet Control Message Protocol (ICMP)
 *
 * Version:	@(#)icmp.c	1.0.2	04/21/93
 *
 * Authors:	Ross Biro, <bir7@leland.Stanford.Edu>
 *		Fred N. van Kempen, <waltje@uWalt.NL.Mugnet.ORG>
 *		Mark Evans, <evansmp@uhura.aston.ac.uk>
 *
 *		This program is free software; you can redistribute it and/or
 *		modify it under the terms of the GNU General Public License
 *		as published by the Free Software Foundation; either version
 *		2 of the License, or (at your option) any later version.
 */
#include <linux/types.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/fcntl.h>
#include <linux/socket.h>
#include <linux/in.h>
#include "inet.h"
#include "timer.h"
#include "dev.h"
#include "ip.h"
#include "route.h"
#include "protocol.h"
#include "icmp.h"
#include "tcp.h"
#include "skbuff.h"
#include "sock.h"
#include <linux/errno.h>
#include <linux/timer.h>
#include <asm/system.h>
#include <asm/segment.h>


#undef	ICMP_DEBUG
#ifdef	ICMP_DEBUG
#   define PRINTK(x)	printk x
#else
#   define PRINTK(x)	/**/
#endif


#define min(a,b)	((a)<(b)?(a):(b))


/* An array of errno for error messages from dest unreach. */
struct icmp_err icmp_err_convert[] = {
  { ENETUNREACH,	1 }, { EHOSTUNREACH,	1 },
  { ENOPROTOOPT,	1 }, { ECONNREFUSED,	1 },
  { EOPNOTSUPP,		0 }, { EOPNOTSUPP,	0 },
  { ENETUNREACH,	1 }, { EHOSTDOWN,	1 },
  { ENONET,		1 }, { ENETUNREACH,	1 },
  { EHOSTUNREACH,	1 }, { EOPNOTSUPP,	0 },
  { EOPNOTSUPP,		0 }
};


/* Display the contents of an ICMP header. */
void
print_icmph(struct icmp_header *icmph)
{
#if ICMP_DEBUG
 printk("ICMP: type = %d, code = %d, checksum = %X\n", icmph->type,
					icmph->code, icmph->checksum);
 printk(" gateway = %X\n", icmph->un.gateway);
#endif
}


/* Send an ICMP message in response to a packet. */
void
icmp_reply(struct sk_buff *skb_in, int type, int code, struct device *dev)
{
  struct sk_buff *skb;
  struct ip_header *iph;
  int offset;
  struct icmp_header *icmph;
  int len;

  PRINTK (("icmp_reply (skb_in = %X, type = %d, code = %d, dev=%X)\n",
	   skb_in, type, code, dev));

  /* Get some memory for the reply. */
  len = dev->hard_header_len +
	sizeof(struct ip_header) + 32 +
	sizeof(struct icmp_header) +
	sizeof(*skb) + 8;	/* amount of header to return. */
	   
  skb = kmalloc(len, GFP_ATOMIC);
  if (skb == NULL) return;

  skb->lock = 0;
  skb->mem_addr = skb;
  skb->mem_len = len;

  len -= sizeof(*skb);

  /* Find the IP header. */
  iph = (struct ip_header *) (skb_in + 1);
  iph = (struct ip_header *) ((unsigned char *) iph + dev->hard_header_len);

  /* Build Layer 2-3 headers for message back to source. */
  offset = ip_build_header(skb, iph->daddr, iph->saddr,
			   &dev, IPPROTO_ICMP, NULL, len);
  if (offset < 0) {
	skb->sk = NULL;
	kfree_skb(skb, FREE_READ);
	return;
  }

  /* Re-adjust length according to actual IP header size. */
  skb->len = offset + sizeof(struct icmp_header) + 8;
  icmph = (struct icmp_header *) ((unsigned char *) (skb + 1) + offset);
  icmph->type = type;
  icmph->code = code;
  icmph->checksum = 0; /* we don't need to compute this. */
  icmph->un.gateway = 0; /* might as well 0 it. */
  memcpy(icmph + 1, iph + 1, 8);

  /* Send it and free it. */
  ip_queue_xmit(NULL, dev, skb, 1);
}


/* Deal with incoming ICMP packets. */
int
icmp_rcv(struct sk_buff *skb1, struct device *dev, struct options *opt,
	 unsigned long daddr, unsigned short len,
	 unsigned long saddr, int redo, struct inet_protocol *protocol)
{
  int size, offset;
  struct icmp_header *icmph, *icmphr;
  struct sk_buff *skb;
  unsigned char *buff;

  /* Drop broadcast packets. */
  if (chk_addr(daddr) == IS_BROADCAST) {
	printk("ICMP: Discarded broadcast from %s\n", in_ntoa(saddr));
	skb1->sk = NULL;
	kfree_skb(skb1, FREE_READ);
	return(0);
  }

  buff = skb1->h.raw;
  icmph = (struct icmp_header *) buff;

  /* Validate the packet first */
  if (icmph->checksum) {	/* Checksums Enabled? */
	if (ip_compute_csum((unsigned char *) icmph, len)) {
		/* Failed checksum! */
		PRINTK (("ICMP failed checksum!\n"));
		skb1->sk = NULL;
		kfree_skb(skb1, FREE_READ);
		return(0);
	}
  }
  print_icmph(icmph);

  /* Parse the ICMP message */
  switch (icmph->type) {
	case ICMP_DEST_UNREACH:
	case ICMP_SOURCE_QUENCH:
		{
			struct ip_header *iph;
			struct inet_protocol *ipprot;
			unsigned char hash;
			int err;

			err = icmph->type << 8 | icmph->code;

			/*
			 * We need to cause the socket to be closed and
			 * the error message to be set appropriately.
			 */
			iph = (struct ip_header *) (icmph + 1);

			/* get the protocol(s) */
			hash = iph->protocol & (MAX_INET_PROTOS -1);

			/* This can change while we are doing it. */
			ipprot = (struct inet_protocol *)inet_protos[hash];
			while (ipprot != NULL) {
				struct inet_protocol *nextip;

				nextip = (struct inet_protocol *)ipprot->next;

				/* Pass it off to everyone who wants it. */
				if (iph->protocol == ipprot->protocol &&
				    ipprot->err_handler)
					ipprot->err_handler(err,
						(unsigned char *)(icmph + 1),
						iph->daddr, iph->saddr, ipprot);
				ipprot = nextip;
			}
			skb1->sk = NULL;
			kfree_skb(skb1, FREE_READ);
			return(0);
		}
	case ICMP_REDIRECT:
		{
			/* We need to put a new route in the routing table. */
			struct ip_header *iph;
			unsigned long ip;

			iph = (struct ip_header *)(icmph+1);
			ip = iph->daddr;
			switch(icmph->code & 7) {
			    case 0:	/* NETWORK UNREACHABLE */
				rt_add((RTF_DYNAMIC | RTF_MODIFIED),
					1, ip, icmph->un.gateway, dev);
				break;
			    case 1:	/* HOST UNREACHABLE */
				rt_add((RTF_DYNAMIC | RTF_MODIFIED | RTF_HOST),
					1, ip, icmph->un.gateway, dev);
				break;
			    case 2:	/* PROTOCOL UNREACHABLE */
			    case 3:	/* PORT UNREACHABLE */
			    case 4:	/* FRAGM. WITH DF SET */
			    case 5:	/* SOURCE ROUTE UNREACH */
				printk("ICMP: UNREACH CODE=%d from %s\n",
				    (icmph->code & 7), in_ntoa(saddr));
				break;
			    default:
				printk("ICMP: weird UNREACH CODE=%d from %s\n",
				    (icmph->code & 7), in_ntoa(saddr));
				break;
			}
			skb1->sk = NULL;
			kfree_skb(skb1, FREE_READ);
			return(0);
		}
	case ICMP_ECHO: 
		/*
		 * Allocate an sk_buff response buffer (assume 64 byte
		 * IP header)
		 */
		size = sizeof(struct sk_buff) + dev->hard_header_len + 64 + len;
		skb = kmalloc(size, GFP_ATOMIC);
		if (skb == NULL) {
			skb1->sk = NULL;
			kfree_skb(skb1, FREE_READ);
			return(0);
		}
		skb->sk = NULL;
		skb->lock = 0;
		skb->mem_addr = skb;
		skb->mem_len = size;

		/* Build Layer 2-3 headers for message back to source */
		offset = ip_build_header(skb, daddr, saddr, &dev,
					 IPPROTO_ICMP, opt, len);
		if (offset < 0) {
			/* Problems building header */
			PRINTK (("Could not build IP Header for ICMP ECHO Response\n"));
			kfree_s(skb->mem_addr, skb->mem_len);
			skb1->sk = NULL;
			kfree_skb(skb1, FREE_READ);
			return(0);	/* just toss the received packet */
		}

		/* Re-adjust length according to actual IP header size. */
		skb->len = offset + len;

		/* Build ICMP_ECHO Response message */
		icmphr = (struct icmp_header *) ((char *) (skb + 1) + offset);
		memcpy((char *) icmphr, (char *) icmph, len);
		icmphr->type = ICMP_ECHOREPLY;
		icmphr->code = 0;
		icmphr->checksum = 0;

		if (icmph->checksum) {	/* Calculate Checksum */
			icmphr->checksum = ip_compute_csum((void *)icmphr, len);
		}

		/* Ship it out - free it when done */
		ip_queue_xmit((volatile struct sock *)NULL, dev, skb, 1);
	
		skb1->sk = NULL;
		kfree_skb(skb1, FREE_READ);
		return(0);
	default:
		PRINTK (("Unsupported ICMP type = x%x\n", icmph->type));
		skb1->sk = NULL;
		kfree_skb(skb1, FREE_READ);
		return(0);	/* just toss the packet */
  }

  /* Should be unecessary, but just in case. */
  skb1->sk = NULL;
  kfree_skb(skb1, FREE_READ);
  return(0);	/* just toss the packet */
}


/* Perform any ICMP-related I/O control requests. */
int
icmp_ioctl (volatile struct sock *sk, int cmd, unsigned long arg)
{
  switch(cmd) {
	default:
		return(-EINVAL);
  }
  return(0);
}
