/*
 * INET		An implementation of the TCP/IP protocol suite for the LINUX
 *		operating system.  INET is implemented using the  BSD Socket
 *		interface as the means of communication with the user level.
 *
 *		This file implements the various access functions for the
 *		PROC file system.  It is mainly used for debugging and
 *		statistics.
 *
 * NOTE:	I want the bulk of this code (and its AF_UNIX counterpart)
 *		to disappear into the user-space "netstat" program. This
 *		requires some standard "netstat" data structures to be
 *		defined, which are filled in in this code... -FvK
 *
 * Version:	@(#)proc.c	1.0.2	03/21/93
 *
 * Authors:	Fred N. van Kempen, <waltje@uWalt.NL.Mugnet.ORG>
 *		Gerald J. Heim, <heim@peanuts.informatik.uni-tuebingen.de>
 *
 *		This program is free software; you can redistribute it and/or
 *		modify it under the terms of the GNU General Public License
 *		as published by the Free Software Foundation; either version
 *		2 of the License, or (at your option) any later version.
 */
#include <linux/autoconf.h>
#include <linux/sched.h>
#include <linux/socket.h>
#include <linux/net.h>
#include <linux/un.h>
#include <linux/in.h>
#include <sys/param.h>
#include "inet.h"
#include "timer.h"
#include "dev.h"
#include "ip.h"
#include "protocol.h"
#include "tcp.h"
#include "udp.h"
#include "skbuff.h"
#include "sock.h"
#include "raw.h"


/*
 * Get__netinfo returns the length of that string.
 *
 * KNOWN BUGS
 *  While the states of TCP connections are standard and well-known
 *  the states of UDP or RAW sockets are undefined. However,
 *  ESTABLISHED is also used in UDP an therefore included.
 *  As in get_unix_netinfo, the buffer might be too small. If this
 *  happens, get__netinfo returns only part of the available infos.
 */
static int
get__netinfo(struct proto *pro, char *buffer)
{
  volatile struct sock **s_array;
  volatile struct sock *sp;
  char *pos=buffer;
  int i;
  unsigned long  dest, src;
  unsigned short destp, srcp;
  char *state;

  pos+=sprintf(pos, "%s\nNum\tProto\tSource:Port\tDestination:Port\tState\n",
								pro->name);

  s_array = pro->sock_array;
  for(i = 0; i < SOCK_ARRAY_SIZE; i++) {
	sp = s_array[i];
	while(sp != NULL) {
		dest  = sp->daddr;
		src   = sp->saddr;
		destp = sp->dummy_th.dest;
		srcp  = sp->dummy_th.source;

		/* Since we are Little Endian we need to swap the bytes :-( */
		destp = ntohs(destp);
		srcp  = ntohs(srcp);

		/* Get socketstate in human-readable form. */
		if (pro == &tcp_prot) {
			switch(sp->state) {
				case TCP_ESTABLISHED:
					state = "ESTABLISHED";
					break;
				case TCP_SYN_SENT:
					state = "SYN_SENT";
					break;
				case TCP_SYN_RECV:
					state = "SYN_RECV";
					break;
				case TCP_FIN_WAIT1:
					state = "FIN_WAIT1";
					break;
				case TCP_FIN_WAIT2:
					state = "FIN_WAIT2";
					break;
				case TCP_TIME_WAIT:
					state = "TIME_WAIT";
					break;
				case TCP_CLOSE:
					state = "CLOSE";
					break;
				case TCP_CLOSE_WAIT:
					state = "CLOSE_WAIT";
					break;
				case TCP_LAST_ACK:
					state = "LAST_ACK";
					break;
				case TCP_LISTEN:
					state = "LISTEN";
					break;
				default:
					printk("INET: netinfo: unknown state.\n");
					state = "UNKNOWN";
			}
		} else if (pro == &udp_prot) {
			switch(sp->state) {
				case TCP_ESTABLISHED:
					/* udp also uses this */
					state = "ESTABLISHED";
					break;
				default:
					state = ""; /* UNKNOWN */
			}
		} else if (pro==&raw_prot) {
			state = ""; /* I don't know if RAW uses states at all ...*/
		} else {
			printk("INET: netinfo: unknown protocol.\n");
			state = "";
		}

		pos+=sprintf(pos, " %d :\t%d\t%d.%d.%d.%d:%d\t%d.%d.%d.%d:%d\t\t%s\n",
			i, sp->protocol,
			src       & 0xff,
			(src>>8)  & 0xff,
			(src>>16) & 0xff,
			(src>>24) & 0xff,
			srcp,
			dest       & 0xff,
			(dest>> 8) & 0xff,
			(dest>>16) & 0xff,
			(dest>>24) & 0xff,
			destp, state);

		/* Is place in buffer too rare? then abort. */
		if (pos > buffer+PAGE_SIZE-100) {
			printk("oops, too many %s sockets for netinfo.\n",
					pro->name);
			return(strlen(buffer));
		}

		/*
		 * All sockets with (port mod SOCK_ARRAY_SIZE) = i
		 * are kept in sock_array[i], so we must follow the
		 * 'next' link to get them all.
		 */
		sp = sp->next;
	}
  }
  return(strlen(buffer));
} 


int tcp_get_info(char *buffer)
{
  return get__netinfo(&tcp_prot, buffer);
}


int udp_get_info(char *buffer)
{
  return get__netinfo(&udp_prot, buffer);
}


int raw_get_info(char *buffer)
{
  return get__netinfo(&raw_prot, buffer);
}
